/* storstuff.cpp - handle storage of stuff data
 *
 * USAGE: storstuff datadir maxsize rdsize
 *
 * This program expects to read data on stdin as sent to it by the
 *
 * Signal handling:
 *  USR2 - start new run
 *  USR1 - start new file
 *  TERM - quit
 *  QUIT - quit
 *  INT - quit
 *
 * Marty Olevitch, May '01
 *	c++ version - Oct '03
 */

#include <cstdio>
#include <iostream>
#include <string>
#include <stor.hpp>
#include <signal.h>

char *Datadir;
unsigned long Maxsize;
unsigned long Rdsize;

stor* Stp;
char* Progname;

// signal handlers

// causes us to quit
void
quit(int sig)
{
    printf("%s: Arghhh! got signal %d - quitting.\n", Progname, sig);
    Stp->close();
    exit(0);
}

// newrun - for USR2 - start a new run
void
newrun(int sig)
{
    try {
	Stp->next_run();
    } catch (std::string s) {
	std::cerr << s << "\n";
	exit(1);
    }
    std::string path = Stp->cur_file_path();
    std::cout << Progname
    	<< ": got USR2 - starting new run " << path << std::endl;
}

// newfile - for USR1 - start a new file
void
newfile(int sig)
{
    try {
	long ret = Stp->next_file();
    } catch (std::string s) {
	std::cerr << s << "\n";
	exit(1);
    }
    std::string path = Stp->cur_file_path();
    std::cout << Progname
    	<< ": got USR1 - starting new file " << path << std::endl;
}

int
main(int argc, char *argv[])
{
    char *d;
    int ret;

    Progname = argv[0];

    if (argc != 4) {
	std::cerr << "USAGE: " << Progname << " datadir maxsize rdsize" << std::endl;
	exit(1);
    }

    Datadir = argv[1];
    Maxsize = atol(argv[2]);
    Rdsize = atol(argv[3]);

    std::cerr << Progname
    	<< " " << Datadir << " " << Maxsize << " " << Rdsize << std::endl;

    d = new char [Maxsize];

    signal(SIGUSR2, newrun);
    signal(SIGUSR1, newfile);
    signal(SIGTERM, quit);
    signal(SIGQUIT, quit);
    signal(SIGINT, quit);

    try {
	stor st(Datadir, Maxsize);
	Stp = &st;
	while (1) {
	    std::cin.read(d, Rdsize);
	    int amt = std::cin.gcount();
	    if (amt != Rdsize) {
		std::cerr << Progname << ": bad read, got " << amt
		    << " wanted " << Rdsize << std::endl;
		if (std::cin.eof() || std::cin.bad() || std::cin.fail()) {
		    st.flush();
		    break;
		}
		continue;
	    }
	    ret = st.save_data(d, Rdsize);
	    if (ret == 0) {
		/* OK */
		;
	    } else if (ret == 1) {
		std::cerr << Progname << ": new file " << st.cur_file() << std::endl;
	    } else {
		std::cerr << Progname << ": unknown return save_data() return code"
		    << std::endl;
	    }
	    st.flush();
	}
    }
    catch (std::string e) {
	std::cerr << e << std::endl;
	exit(1);
    }

    return 0;
}
