
/*
 Copyright (C) 2003 Martin A. Olevitch
 Written by Martin A. Olevitch (marty@cosray.wustl.edu)

 This file is part of Stor.

 Stor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 Stor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Stor, see the file COPYING; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* stor.cpp - WU LEXAS data storage library
 *
 *	Marty Olevitch, 1997-2003
 */

#include "stor.hpp"

#include <string>
#include <iomanip>
#include <fstream>

// Some systems don't have sstream yet.
#if defined(HAVE_SSTREAM)
# include <sstream>
# define OSTRINGSTREAM std::ostringstream
# define ISTRINGSTREAM std::istringstream
#elif defined(HAVE_STRSTREAM)
# include <strstream>
# define OSTRINGSTREAM ostrstream
# define ISTRINGSTREAM istrstream
#endif

#if defined(STD_IOS_BASE_APP)
# define APPEND_FLAG std::ios_base::app
#elif defined(IOS_APP)
# define APPEND_FLAG ios::app
#endif

#include <cerrno>
#include <cstring>
#include <sys/stat.h>
#include <unistd.h>	// getpid, open
#include <fcntl.h>	// open modes

#define MKDIR(x, y)	mkdir(x, y)
#define STATERR	ENOENT

using namespace std;

namespace {
    string const CUR_RUN_FILE(".stor_cur_run");
    string const CUR_DATA_FILE(".stor_cur_file");
    string const ACTIVE_FILE(".stor_active");
    string const COMMENT_FILE("stor_comments");

    // open_active_file() - Make sure no other stor instance is
    // using this datadir by checking for the existence of the active file.
    // It will throw an exception if another stor is using it. The active
    // file resides in the datadir and contains the pid of the process that
    // is using it.
    static void
    open_active_file(string const & datadir, string const & active_file)
    {
	int fd = open(active_file.c_str(), O_CREAT | O_EXCL | O_WRONLY, 0644);

	if (fd == -1) {
	    // active file already exists, meaning that some other stor
	    // instance is using the same datadir. Read the pid from the
	    // active file and report it.
	    ifstream af(active_file.c_str());
	    string pid;
	    if (af) {
		af >> pid;
		af.close();
	    } else {
		pid = "????";
	    }
	    OSTRINGSTREAM ost;
	    ost << "stor: another stor instance (process " 
	    	<< pid << ") appears to be using '" << datadir
		<< "' to save data.";
	    throw ost.str();
	}

	// Write our pid into the active file we just opened.
	OSTRINGSTREAM pid;
	pid << getpid() << endl;
	int ret = write(fd, pid.str().c_str(), pid.str().size());
	if (ret != pid.str().size()) {
	    OSTRINGSTREAM ost;
	    ost << "stor: open_active_file: bad write to '"
		<< active_file 
		<< "' (" << strerror(errno) << ")";
	    throw ost.str();
	}
	close(fd);
    }

    // end of anonymous namespace
}

stor::stor(string _datadir, long _maxfilesize, bool _do_default_comment)
    :
	datadir(_datadir),
	maxfilesize(_maxfilesize),
	do_default_comment(_do_default_comment)
{
    if (maxfilesize < 0L) {
	switch_data_files = false;
	maxfilesize = 0L;
    } else {
	switch_data_files = true;
    }

    struct stat statbuf;
    if (stat(datadir.c_str(), &statbuf)) {
	if (errno == STATERR) {
	    // datadir doesn't exist, try to make it.
	    if (MKDIR(datadir.c_str(), 0755)) {
		OSTRINGSTREAM ost;
		ost << "stor: can't create datadir '" << datadir
		    << "' (" << strerror(errno) << ") ";
		throw ost.str();
	    }
	} else {
	    // bad stat
	    OSTRINGSTREAM ost;
	    ost << "stor: bad stat on datadir '" << datadir
		<< "' (" << strerror(errno) << ") ";
	    throw ost.str();
	}
    } else {
	// datadir exists; make sure it is a directory
	if (!S_ISDIR(statbuf.st_mode)) {
	    OSTRINGSTREAM ost;
	    ost << "stor: datadir '" << datadir
	    	<< "' exists, but is not a directory!";
	    throw ost.str();
	}
    }

    // Make sure no other stor instance is using this datadir by
    // checking for the existence of the active_file. It will throw an
    // exception if another stor is using it.
    active_file = datadir + "/" + ACTIVE_FILE;
    open_active_file(datadir, active_file);

    // need to read saved run file to get run number
    cur_run_name_file = datadir + "/" + CUR_RUN_FILE;
    ifstream f(cur_run_name_file.c_str());
    // if f is bad, then we assume no file; use initial run no. of -1.
    if (f) {
	f >> run;
	f.close();
    } else {
	run = -1;
    }

    next_run();

    totbytes = 0L;
    if (do_default_comment) {
	comment("init");
    }
}

/* next_run - increment the run number; make the new run directory; file
 * number will be set to 0. Throws a string as an exception if there is a
 * problem. */
void
stor::next_run()
{
    long tmprun = run + 1;

    OSTRINGSTREAM runost;
    runost << setw(5) << setfill('0') << tmprun;
    rundir = datadir + "/" + runost.str();

    // Create the new run directory.
    if (MKDIR(rundir.c_str(), 0755)) {
	string s = "stor: next_run: can't mkdir '" + rundir + 
		"' (" + strerror(errno) + ")";
	throw s;
    }

    // Write the new run number to the run name file.
    ofstream f(cur_run_name_file.c_str());
    if (f) {
	f << runost.str();
	f.close();
    } else {
	string s = "stor: next_run: can't open '" + cur_run_name_file +
		"' (" + strerror(errno) + ")";
	throw s;
    }

    run++;
    file = -1;

    next_file();
}

// next_file: close current data file; open next one. Returns new data
// file number if successful, else throws a string as an exception.
long
stor::next_file()
{
    if (filestream.is_open()) {
	filestream.close();
    }

    file++;

    OSTRINGSTREAM fileost;
    fileost << setw(6) << setfill('0') << file;

    string filename = rundir + "/" + fileost.str();
    filestream.open(filename.c_str());
    if (!filestream) {
	string s = "stor: next_file: can't open '" + filename +
		"' (" + strerror(errno) + ")";
	file--;
	throw s;
    }

    // update the current file name file.
    string cur_file_name_file = rundir + "/" + CUR_DATA_FILE;
    ofstream f(cur_file_name_file.c_str());
    if (f) {
	f << fileost.str();
	f.close();
    } else {
	string s = "stor: next_file: can't open '" + cur_file_name_file +
		"' (" + strerror(errno) + ")";
	file--;
	throw s;
    }
    filesize = 0;
    return file;
}

void
stor::flush()
{
    if (!filestream.flush()) {
	OSTRINGSTREAM ost;
	ost <<  "stor: next_file: error: " <<  strerror(errno);
	throw ost.str();
    }
}

// save_data: write the data to the current file. Throws a string as an
// exception on error, 1 if we started a new data file, else 0.
// dp = pointer to data, amtb = size in bytes of data.
int
stor::save_data(char const *dp, int amtb)
{
    if (amtb == 0) {
	return 0;
    }

    int retcode = 0;

    if (amtb > maxfilesize) {
	OSTRINGSTREAM ost;
	ost << "stor: save_data: amtb (" << amtb << ") too big";
	throw ost.str();
    }

    if (switch_data_files) {
	if (filesize + amtb > maxfilesize) {
            next_file();
	    retcode = 1;
	}
    }

    if (!filestream.write(dp, amtb)) {
	OSTRINGSTREAM ost;
	ost << "stor: save_data: bad write: " << strerror(errno);
	throw ost.str();
    }

    filesize += amtb;
    totbytes += amtb;
    return retcode;
}

void
stor::comment(string const & s)
{
    comment(s.c_str());
}

void
stor::comment(char const *s)
{
    // Get time.
    time_t t = time(NULL);
    string timestr = ctime(&t);
    timestr[timestr.size()-1] = '\0';	// delete \n

    // Enter comment in the datadir comment file.
    string filename = datadir + "/" + COMMENT_FILE;
    ofstream f(filename.c_str(), APPEND_FLAG);
    f << timestr << ": " << rundir << ": " << s << endl;
    f.close();

    // Enter comment in the run's comment file.
    filename = rundir + "/" + COMMENT_FILE;
    f.open(filename.c_str(), APPEND_FLAG);
    f << timestr << ": " << rundir << ": " << s << endl;
    f.close();
}

void
stor::close()
{
    if (do_default_comment) {
	comment("close");
    }
    unlink(active_file.c_str());
    filestream.close();
}

stor::~stor()
{
    close();
}

long stor::cur_run() const { return run; }
long stor::cur_file() const { return file; }
long stor::bytes_cur_file() const { return filesize; }
long stor::total_bytes() const { return totbytes; }

string
stor::cur_file_str() const
{
    OSTRINGSTREAM ost;
    ost << setw(6) << setfill('0') << file;
    return string(ost.str());
}

string
stor::cur_run_str() const
{
    OSTRINGSTREAM ost;
    ost << setw(5) << setfill('0') << run;
    return string(ost.str());
}

string
stor::cur_file_path() const
{
    OSTRINGSTREAM ost;
    ost << rundir << "/" << setw(6) << setfill('0') << file;
    return string(ost.str());
}

string
stor::version()
{
    string s = STOR_VERSION;
    return s;
}

#ifdef TESTMAIN
#include <iostream>
int
main(int argc, char *argv[])
{
    cout << "stor version " << stor::version() << endl;

    try {
	stor st("/barfola", 100);
    }
    catch (string e) {
	cerr << e << endl;
    }

    try {
	stor st("foo", 25);
	string s;
	st.comment("This is a char * comment");
	cout << "current run is  " << st.cur_run() << endl;
	cout << "current file is " << st.cur_file() << endl;
	while (cin >> s) {
	    st.save_data(s.c_str(), s.size());
	    cout << "bytes in file: " << st.bytes_cur_file()
	    	<< "     total bytes: " << st.total_bytes() << endl;
	}
	s = "This is a string comment";
	st.comment(s);
	cout << "current file: " << st.cur_file_str() << endl;
	cout << "current run: " << st.cur_run_str() << endl;
	cout << "current path: " << st.cur_file_path() << endl;
    }
    catch (string e) {
	cerr << e << endl;
	exit(1);
    }
}
#endif // TESTMAIN
