/* stor.c - WU LEXAS data storage library
 *
 *	Marty Olevitch, July 1997
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "stor.h"

#ifdef __BORLANDC__
#define STOR_VERSION "1.4a" 
#include <dir.h>	/* for mkdir() */
# define MKDIR(x, y)	mkdir(x)
# define STATERR	ENOFILE
# define S_ISDIR(m)	((m) & S_IFDIR)
# define CUR_RUN_FILE	"_cur_run"
#else
# define MKDIR(x, y)	mkdir(x, y)
# define STATERR	ENOENT
# define CUR_RUN_FILE	".stor_cur_run"
#endif

#ifndef HAVE_STRERROR
char *strerror(int e);
#endif

static long Currun;		/* current run number */
static long Curfile;		/* current file number */
static long Curfilesize;	/* size (bytes) of current file */
static FILE *Curfp;
static char *Datadir;		/* root directory of the data */
static char Errmesg[100];	/* error message */
static long Maxfilesize;	/* max size of each data file */
static char *Rundir;		/* root directory of the current run */
static int Switch_data_files;	/* zero if we should put all the data in
				 * one file */
static unsigned long Totalbyteswrit;

static char *Dirstr;		/* utility string */

static int Init_init = 0;
static int Init_next_file = 0;
static int Init_next_run = 0;

int
stor_init(datadir, maxfilesize)
    char *datadir;	/* name of root data directory */
    long maxfilesize;	/* max size of each file; if zero don't switch files */
{
    int ret;
    struct stat statbuf;

    if (Init_init) {
	sprintf(Errmesg, "stor_init: already initialized");
	return -1;
    }

    Rundir = NULL;
    if (maxfilesize <= 0L) {
	Switch_data_files = 0;
	Maxfilesize = 0L;
    } else {
	Switch_data_files = 1;
	Maxfilesize = maxfilesize;
    }

    ret = stat(datadir, &statbuf);
    if (ret) {
	if (errno == STATERR) {
	    /* datadir doesn't exist; make one */
	    ret = MKDIR(datadir, 0755);
	    if (ret) {
		sprintf(Errmesg, "stor_init: can't create datadir '%s' (%s)", 
		    datadir, strerror(errno));
		return -1;
	    }
	} else {
	    /* bad stat */
	    sprintf(Errmesg, "stor_init: bad stat on '%s' (%s)", datadir,
	    	strerror(errno));
	    return -1;
	}
    } else {
	/* datadir exists; make sure it is a directory */
	if (!S_ISDIR(statbuf.st_mode)) {
	    sprintf(Errmesg,"Datadir '%s' exists but is not a directory!",
		datadir);
	    return -1;
	}
    }

    Datadir = malloc(strlen(datadir)+1);
    if (Datadir == NULL) {
	sprintf(Errmesg, "stor_init: can't malloc Datadir string");
	return -1;
    }
    strcpy(Datadir, datadir);

    Dirstr = malloc(strlen(datadir) + 50);
    if (Dirstr == NULL) {
	sprintf(Errmesg, "stor_init: can't malloc Dirstr string");
	return -1;
    }

    Init_init = 1;
    Currun = -1;
    ret = stor_next_run();	/* stor_next_run() calls stor_next_file()
				 * to set up the initial file number */
    if (ret == -1) {
	free(Datadir);
	free(Dirstr);
	Init_init = 0;
	return -1;
    }

    Totalbyteswrit = 0L;

    return 0;
}

/* stor_next_run - increment the run number; make the new run directory;
 * file number will be set to 0. Returns 0 if all went well, else -1. */
int
stor_next_run(void)
{
    FILE *fp;
    int ret;
    char *retp;
    char val[20];

    if (!Init_init) {
	sprintf(Errmesg,"stor_next_run: not initialized - run stor_init()");
	return -1;
    }

    sprintf(Dirstr, "%s/%s", Datadir, CUR_RUN_FILE);

    if (Currun == -1) {
	/* need to read saved run file to get run number */
	fp = fopen(Dirstr, "r");
	/* if fp == NULL, we assume no file; use default initial run number */
	if (fp != NULL) {
	    retp = fgets(val, 20, fp);
	    if (retp == NULL) {
		sprintf(Errmesg, "stor_next_run: can't read '%s' (%s)",
		    Dirstr, strerror(errno));
		fclose(fp);
		return -1;
	    }
	    Currun = atoi(val);
	    fclose(fp);
	}
    }

    Currun++;

    if (Rundir == NULL) {
	/* first time */
	Rundir = malloc(strlen(Datadir) + 20);
	if (Rundir == NULL) {
	    sprintf(Errmesg, "stor_next_run: can't malloc Rundir");
	    Currun--;
	    return -1;
	}
    }
    sprintf(Rundir, "%s/%05d", Datadir, Currun);
    ret = MKDIR(Rundir, 0755);
    if (ret) {
	sprintf(Errmesg, "stor_next_run: can't mkdir '%s' (%s)", Rundir,
	    strerror(errno));
	Currun--;
	return -1;
    }

    /* now write the new run number to the saved run file */
    fp = fopen(Dirstr, "w");
    if (fp == NULL) {
	sprintf(Errmesg,"stor_next_run: can't fopen(\"%s\", \"w\") (%s)",
	    Dirstr, strerror(errno));
	Currun--;
	return -1;
    }
    fprintf(fp, "%05d", Currun);
    fclose(fp);

    Init_next_run = 1;

    Curfile = -1;
    ret = stor_next_file();
    if (ret == -1) {
	free(Rundir);
	Init_next_run = 0;
	return -1;
    }


    return 0;
}

/* stor_next_file: close current data file; open next one. Returns new data
 * file number if successful, else -1. */
int
stor_next_file(void)
{
    if (!Init_next_run) {
	sprintf(Errmesg, "stor_next_file: not initialized. Run stor_init()");
	return -1;
    }

    if (Curfp != NULL) {
	fclose(Curfp);
    }

    Curfile++;
    sprintf(Dirstr, "%s/%06d", Rundir, Curfile);
    Curfp = fopen(Dirstr, "wb");
    if (Curfp == NULL) {
	sprintf(Errmesg, "stor_next_file: can't open '%s' (%s)", Dirstr,
	    strerror(errno));
	Curfile--;
	return -1;
    }

    Init_next_file = 1;
    Curfilesize = 0;
    return Curfile;
}

/* stor_save_data: write the data to the current file. Returns -1 on error,
 * 1 if we started a new data file, else 0. */
int
stor_save_data(dp, size, nitems)
    unsigned char *dp;	/* pointer to data */
    size_t size;	/* size of each data item */
    size_t nitems;	/* no. of data items */
{
    size_t amt;
    int ret;
    int retcode;

    if (!Init_next_run) {
	sprintf(Errmesg, "stor_save_data: not initialized. Run stor_init()");
	return -1;
    }

    if (nitems == 0) {
	return 0;
    }
    
    retcode = 0;
    amt = size * nitems;
    if (amt > Maxfilesize) {
	sprintf(Errmesg, "stor_save_data: amt (%d) too big", amt);
	return -1;
    }
    if (Switch_data_files) {
	if (Curfilesize + amt > Maxfilesize) {
	    ret = stor_next_file();
	    if (ret == -1) {
		return -1;
	    }
	    retcode = 1;
	}
    }

    if (Curfp == NULL) {
	sprintf(Errmesg, "stor_save_data: Curfp is NULL");
	return -1;
    }

    ret = fwrite(dp, size, nitems, Curfp);
    if (ret != nitems) {
	if (ferror(Curfp)) {
	    Curfp = NULL;
	    sprintf(Errmesg, "stor_save_data: bad fwrite (%s)",
		strerror(errno));
	    return -1;
	} else if (feof(Curfp)) {
	    sprintf(Errmesg,
	    	"stor_save_data: EOF on output '%s/%06d', shouldn't happen", 
		Rundir, Curfile);
	    Curfp = NULL;
	    return -1;
	} else {
	    sprintf(Errmesg, "stor_save_data: bad return from fwrite (%s)",
	    	strerror(errno));
	    return -1;
	}
    }

    Curfilesize += amt;
    Totalbyteswrit += amt;
    Errmesg[0] = '\0';
    return retcode;
}

int
stor_get_cur_run(void)
{
    return Currun;
}

int
stor_get_cur_file(void)
{
    return Curfile;
}

int
stor_bytes_cur_file(void)
{
    return Curfilesize;
}

int
stor_total_bytes(void)
{
    return Totalbyteswrit;
}

char *
stor_strerror(void)
{
    return Errmesg;
}

void
stor_get_cur_run_str(s)
    char *s;
{
    sprintf(s, "%05d", Currun);
}

void
stor_get_cur_file_str(s)
    char *s;
{
    sprintf(s, "%06d", Curfile);
}

void
stor_cur_file_path(s)
    char *s;
{
    sprintf(s, "%s/%06d", Rundir, Curfile);
}

void
stor_close(void)
{
    if (!Init_init) {
	return;
    }

    if (Curfp != NULL) {
	fclose(Curfp);
	Curfp = NULL;
    }
    free(Datadir);
    free(Dirstr);
    free(Rundir);
    Init_init = Init_next_run = Init_next_file = 0;
}

void
stor_version(char *s)
{
    strcpy(s, STOR_VERSION);
}

#ifndef HAVE_STRERROR
/* strerror - the equivalent of the ANSI strerror, I hope
 *
 * Marty Olevitch, Feb 1996 */

extern int sys_nerr;
extern char *sys_errlist[];

char *
strerror(int e)
{
    if (e < sys_nerr) {
	return(sys_errlist[e]);
    } else {
	return NULL;
    }
}
#endif /* not HAVE_STRERROR */
