
/*
 Copyright (C) 2003 Martin A. Olevitch
 Written by Martin A. Olevitch (marty@cosray.wustl.edu)

 This file is part of Stor.

 Stor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 Stor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Stor, see the file COPYING; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* storstuff.c - handle storage of data
 *
 * USAGE: storstuff datadir maxsize rdsize
 *
 * This program expects to read data on stdin as sent to it by the
 * datad program. It saves the data using the stor library routines.
 *
 * Signal handling:
 *  USR2 - start new run
 *  USR1 - start new file
 *  TERM - quit
 *  QUIT - quit
 *  INT - quit
 *
 * Marty Olevitch, May '01
 */

#include <stdio.h>
#include <stor.h>
#include <signal.h>
#include <stdlib.h>

char *Datadir;
unsigned long Maxsize;
char *Progname;
unsigned long Rdsize;

// signal handlers

// causes us to quit
void
quit(int sig)
{
    printf("%s: Arghhh! got signal %d - quitting.\n", Progname, sig);
    stor_flush();
    stor_close();
    exit(0);
}

// newrun - for USR2 - start a new run
void
newrun(int sig)
{
    char path[1024];
    int ret = stor_next_run();
    if (ret) {
	printf("%s: bad stor_next_run (%s)\n", Progname, stor_strerror());
	fflush(stdout);
	return;
    }
    stor_cur_file_path(path);
    printf("%s: got USR2 - starting new run %s\n", Progname, path );
    fflush(stdout);
}

// newfile - for USR1 - start a new file
void
newfile(int sig)
{
    char path[1024];
    long ret = stor_next_file();
    if (ret < 0) {
	printf("%s: bad stor_next_file (%s)\n", Progname, stor_strerror());
	fflush(stdout);
	return;
    }
    stor_cur_file_path(path);
    printf("%s: got USR1 - starting new file %s\n", Progname, path );
    fflush(stdout);
}

int
main(int argc, char *argv[])
{
    size_t amt;
    unsigned char *d;
    int ret;

    Progname = argv[0];

    if (argc != 4) {
	fprintf(stderr, "USAGE: %s datadir maxsize rdsize\n", Progname);
	exit(1);
    }

    Datadir = argv[1];
    Maxsize = atol(argv[2]);
    Rdsize = atol(argv[3]);

    fprintf(stderr, "%s %s %lu %lu\n", Progname, Datadir, Maxsize, Rdsize);
    fflush(stderr);

    d = (unsigned char *)malloc(Maxsize);
    if (d == NULL) {
	fprintf(stderr, "%s: can't malloc d\n", Progname);
	exit(1);
    }

    ret = stor_init(Datadir, Maxsize);
    if (ret) {
	fprintf(stderr, "%s: bad stor_init (%s). Exiting.\n",
	    Progname, stor_strerror());
	exit(1);
    }

    signal(SIGUSR2, newrun);
    signal(SIGUSR1, newfile);
    signal(SIGTERM, quit);
    signal(SIGQUIT, quit);
    signal(SIGINT, quit);

    while (1) {
	amt = fread(d, 1, Rdsize, stdin);
	if (amt != Rdsize) {
	    fprintf(stderr, "%s: bad read, got %ld wanted %d\n",
	    	Progname, amt, Rdsize);
	    if (feof(stdin) || ferror(stdin)) {
		stor_flush();
		exit(0);
	    }
	    continue;
	}
	ret = stor_save_data(d, sizeof(char), Rdsize);
	if (ret == 0) {
	    /* OK */
	    ;
	} else if (ret == 1) {
	    fprintf(stderr, "%s: new file %d\n", Progname, stor_get_cur_file());
	} else {
	    fprintf(stderr, "%s: error saving data (%s)\n",
		Progname, stor_strerror());
	}
	stor_flush();
    }
}
