
/*
 Copyright (C) 2003 Martin A. Olevitch
 Written by Martin A. Olevitch (marty@cosray.wustl.edu)

 This file is part of Stor.

 Stor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 Stor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Stor, see the file COPYING; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* stor.c - WU LEXAS data storage library
 *
 *	Marty Olevitch, July 1997
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include "stor.h"

#ifdef __BORLANDC__
#define STOR_VERSION "2.01" 
#include <dir.h>	// for mkdir()
# define MKDIR(x, y)	mkdir(x)
# define STATERR	ENOFILE
# define S_ISDIR(m)	((m) & S_IFDIR)
# define CUR_RUN_FILE	"_cur_run"
# define CUR_DATA_FILE	"_cur_file"
#else
# define MKDIR(x, y)	mkdir(x, y)
# define STATERR	ENOENT
# define CUR_RUN_FILE	".stor_cur_run"
# define CUR_DATA_FILE	".stor_cur_file"
#endif

#define COMMENT_FILE	"stor_comments"

#ifndef HAVE_STRERROR
char *strerror(int e);
#endif

static long Currun = -1;	// current run number
static long Curfile;		// current file number
static long Curfilesize;	// size (bytes) of current file
static FILE *Curfp;
static char *Datadir;		// root directory of the data
static int Do_default_comment = 1;
static char Errmesg[100];	// error message
static long Maxfilesize;	// max size of each data file
static char *Rundir;		// root directory of the current run
static int Switch_data_files;	// zero if we should put all the data in
				// one file
static unsigned long Totalbyteswrit;

static char *Dirstr;		// utility string

static int Init_init = 0;
static int Init_next_file = 0;
static int Init_next_run = 0;

// stor_init 
// Returns 0 if all went well, else -1.
// datadir     - name of root data directory 
// maxfilesize - max size of each file; if zero don't switch files 
int
stor_init( char *datadir, long maxfilesize)
{
    int ret;
    struct stat statbuf;

    if (Init_init) {
	sprintf(Errmesg, "stor_init: already initialized");
	return -1;
    }

    Rundir = NULL;
    if (maxfilesize <= 0L) {
	Switch_data_files = 0;
	Maxfilesize = 0L;
    } else {
	Switch_data_files = 1;
	Maxfilesize = maxfilesize;
    }

    ret = stat(datadir, &statbuf);
    if (ret) {
	if (errno == STATERR) {
	    // datadir doesn't exist; make one
	    ret = MKDIR(datadir, 0755);
	    if (ret) {
		sprintf(Errmesg, "stor_init: can't create datadir '%s' (%s)", 
		    datadir, strerror(errno));
		return -1;
	    }
	} else {
	    // bad stat
	    sprintf(Errmesg, "stor_init: bad stat on '%s' (%s)", datadir,
	    	strerror(errno));
	    return -1;
	}
    } else {
	// datadir exists; make sure it is a directory
	if (!S_ISDIR(statbuf.st_mode)) {
	    sprintf(Errmesg,"Datadir '%s' exists but is not a directory!",
		datadir);
	    return -1;
	}
    }

    Datadir = malloc(strlen(datadir)+1);
    if (Datadir == NULL) {
	sprintf(Errmesg, "stor_init: can't malloc Datadir string");
	return -1;
    }
    strcpy(Datadir, datadir);

    Dirstr = malloc(strlen(datadir) + 50);
    if (Dirstr == NULL) {
	sprintf(Errmesg, "stor_init: can't malloc Dirstr string");
	return -1;
    }

    Init_init = 1;
    ret = stor_next_run();	// stor_next_run() calls stor_next_file()
				// to set up the initial file number
    if (ret == -1) {
	free(Datadir);
	free(Dirstr);
	Init_init = 0;
	return -1;
    }

    Totalbyteswrit = 0L;

    if (Do_default_comment) {
	stor_comment("init\n");
    }
    return 0;
}

void
stor_set_cur_run(long r)
{
    Currun = r - 1;
}

// stor_next_run - increment the run number; make the new run directory;
// file number will be set to 0. Returns 0 if all went well, else -1.
int
stor_next_run(void)
{
    FILE *fp;
    int ret;
    char *retp;
    char val[20];

    if (!Init_init) {
	sprintf(Errmesg,"stor_next_run: not initialized - run stor_init()");
	return -1;
    }

    sprintf(Dirstr, "%s/%s", Datadir, CUR_RUN_FILE);

    if (Currun == -1) {
	// need to read saved run file to get run number
	fp = fopen(Dirstr, "r");
	// if fp == NULL, we assume no file; use default initial run number
	if (fp != NULL) {
	    retp = fgets(val, 20, fp);
	    if (retp == NULL) {
		sprintf(Errmesg, "stor_next_run: can't read '%s' (%s)",
		    Dirstr, strerror(errno));
		fclose(fp);
		return -1;
	    }
	    Currun = atoi(val);
	    fclose(fp);
	}
    }

    Currun++;

    if (Rundir == NULL) {
	// first time
	Rundir = malloc(strlen(Datadir) + 20);
	if (Rundir == NULL) {
	    sprintf(Errmesg, "stor_next_run: can't malloc Rundir");
	    Currun--;
	    return -1;
	}
    }
    sprintf(Rundir, "%s/%05d", Datadir, Currun);
    ret = MKDIR(Rundir, 0755);
    if (ret) {
	sprintf(Errmesg, "stor_next_run: can't mkdir '%s' (%s)", Rundir,
	    strerror(errno));
	Currun--;
	return -1;
    }

    // now write the new run number to the saved run file
    fp = fopen(Dirstr, "w");
    if (fp == NULL) {
	sprintf(Errmesg,"stor_next_run: can't fopen(\"%s\", \"w\") (%s)",
	    Dirstr, strerror(errno));
	Currun--;
	return -1;
    }
    fprintf(fp, "%05d\n", Currun);
    fclose(fp);

    Init_next_run = 1;

    Curfile = -1;
    ret = stor_next_file();
    if (ret == -1) {
	free(Rundir);
	Init_next_run = 0;
	return -1;
    }


    return 0;
}

// stor_next_file: close current data file; open next one. Returns new data
// file number if successful, else -1.
long
stor_next_file(void)
{
    if (!Init_next_run) {
	sprintf(Errmesg, "stor_next_file: not initialized. Run stor_init()");
	return -1;
    }

    if (Curfp != NULL) {
	fclose(Curfp);
    }

    Curfile++;
    sprintf(Dirstr, "%s/%06d", Rundir, Curfile);
    Curfp = fopen(Dirstr, "wb");
    if (Curfp == NULL) {
	sprintf(Errmesg, "stor_next_file: can't open '%s' (%s)", Dirstr,
	    strerror(errno));
	Curfile--;
	return -1;
    }

    {
	FILE *fp;
	sprintf(Dirstr, "%s/%s", Rundir, CUR_DATA_FILE);
	fp = fopen(Dirstr, "w");
	if (fp == NULL) {
	    sprintf(Errmesg, "stor_next_file: can't open '%s' (%s)",
	    	Dirstr, strerror(errno));
	    Curfile--;
	    return -1;
	}
	fprintf(fp, "%06d\n", Curfile);
	fclose(fp);
    }

    Init_next_file = 1;
    Curfilesize = 0;
    return Curfile;
}

int
stor_set_stdio_buf_size(size_t sizeb)
{
    int ret;
    static unsigned char *Stdio_buf = NULL;

    free(Stdio_buf);

    Stdio_buf = malloc(sizeb);
    if (Stdio_buf == NULL) {
	return -1;
    }

    ret = setvbuf(Curfp, Stdio_buf, _IOFBF, sizeb);
    return ret;
}

int
stor_flush(void)
{
    int ret;

    if (!Init_next_run) {
	sprintf(Errmesg, "stor_save_data: not initialized. Run stor_init()");
	return -1;
    }

    if (Curfp == NULL) {
	sprintf(Errmesg, "stor_flush: Curfp is NULL");
	return -1;
    }

    ret = fflush(Curfp);
    if (ret) {
	sprintf(Errmesg, "stor_flush: bad fflush (%s)", strerror(errno));
	return -1;
    }

    return 0;
}


// stor_save_data: write the data to the current file.
// Returns -1 on error, 1 if we started a new data file, else 0. 
// dp     - pointer to data 
// size   - size in bytes of each data item 
// nitems - no. of data items 
int
stor_save_data(unsigned char *dp, size_t size, size_t nitems)
{
    size_t amt;
    int ret;
    int retcode;

    if (!Init_next_run) {
	sprintf(Errmesg, "stor_save_data: not initialized. Run stor_init()");
	return -1;
    }

    if (nitems == 0) {
	return 0;
    }
    
    retcode = 0;
    amt = size * nitems;
    if (Switch_data_files && amt > Maxfilesize) {
	sprintf(Errmesg, "stor_save_data: amt (%d) too big", amt);
	return -1;
    }
    if (Switch_data_files) {
	if (Curfilesize + amt > Maxfilesize) {
	    ret = stor_next_file();
	    if (ret == -1) {
		return -1;
	    }
	    retcode = 1;
	}
    }

    if (Curfp == NULL) {
	sprintf(Errmesg, "stor_save_data: Curfp is NULL");
	return -1;
    }

    ret = fwrite(dp, size, nitems, Curfp);
    if (ret != nitems) {
	if (ferror(Curfp)) {
	    Curfp = NULL;
	    sprintf(Errmesg, "stor_save_data: bad fwrite (%s)",
		strerror(errno));
	    return -1;
	} else if (feof(Curfp)) {
	    sprintf(Errmesg,
	    	"stor_save_data: EOF on output '%s/%06d', shouldn't happen", 
		Rundir, Curfile);
	    Curfp = NULL;
	    return -1;
	} else {
	    sprintf(Errmesg, "stor_save_data: bad return from fwrite (%s)",
	    	strerror(errno));
	    return -1;
	}
    }

    Curfilesize += amt;
    Totalbyteswrit += amt;
    Errmesg[0] = '\0';
    return retcode;
}

long
stor_get_cur_run(void)
{
    return Currun;
}

long
stor_get_cur_file(void)
{
    return Curfile;
}

long
stor_bytes_cur_file(void)
{
    return Curfilesize;
}

unsigned long
stor_total_bytes(void)
{
    return Totalbyteswrit;
}

char *
stor_strerror(void)
{
    return Errmesg;
}

void
stor_get_cur_run_str(char *s)
{
    sprintf(s, "%05d", Currun);
}

void
stor_get_cur_file_str(char *s)
{
    sprintf(s, "%06d", Curfile);
}

void
stor_cur_file_path(char *s)
{
    sprintf(s, "%s/%06d", Rundir, Curfile);
}

void
stor_close(void)
{
    if (!Init_init) {
	return;
    }

    if (Curfp != NULL) {
	fclose(Curfp);
	Curfp = NULL;
    }
    if (Do_default_comment) {
	stor_comment("close\n");
    }
    free(Datadir);
    free(Dirstr);
    free(Rundir);
    Init_init = Init_next_run = Init_next_file = 0;
}

void
stor_version(char *s)
{
    strcpy(s, STOR_VERSION);
}

void
stor_no_default_comment(void)
{
    Do_default_comment = 0;
}

int
stor_comment(char *s)
{
    char timestr[64];
    FILE *fp;
    time_t t = time(NULL);
    sprintf(timestr, "%s", ctime(&t));
    timestr[strlen(timestr)-1] = '\0';	// delete \n

    // Enter the comment in the datadir comment file.
    sprintf(Dirstr, "%s/%s", Datadir, COMMENT_FILE);
    fp = fopen(Dirstr, "a");
    if (fp != NULL) {
	fprintf(fp, "%s: ", timestr);
	fprintf(fp, "%s: %s", Rundir, s);
	fclose(fp);
    }

    // Enter the comment in the run's comment file.
    sprintf(Dirstr, "%s/%s", Rundir, COMMENT_FILE);
    fp = fopen(Dirstr, "a");
    if (fp != NULL) {
	fprintf(fp, "%s: ", timestr);
	fprintf(fp, "%s: %s", Rundir, s);
	fclose(fp);
    }
}

#ifndef HAVE_STRERROR
// strerror - the equivalent of the ANSI strerror, I hope
 *
 * Marty Olevitch, Feb 1996

extern int sys_nerr;
extern char *sys_errlist[];

char *
strerror(int e)
{
    if (e < sys_nerr) {
	return(sys_errlist[e]);
    } else {
	return NULL;
    }
}
#endif // not HAVE_STRERROR
